(ns com.aurellem.run.bootstrap-1
  (:use (com.aurellem.gb saves gb-driver util
                         items vbm characters money))
  (:use (com.aurellem.run util title save-corruption bootstrap-0))
  (:use (com.aurellem.exp item-bridge))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn pc-item-writer-program
  []
  (let [limit 201
        [target-high target-low] (disect-bytes-2 0xD162)]
    (flatten
     [[0x00  ;; (item-hack) set increment stack pointer no-op
       0x1E  ;; load limit into E
       limit
       0x3F  ;; (item-hack) set carry flag no-op

       ;; load 2 into C.
       0x0E   ;; C == 1 means input-first nybble
       0x04   ;; C == 0 means input-second nybble

       0x21 ;; load target into HL
       target-low
       target-high
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0x37 ;; (item-hack) set carry flag no-op
       
       0x00 ;; (item-hack) no-op
       0xF3 ;; disable interrupts
       ;; Input Section

       0x3E ;; load 0x20 into A, to measure buttons
       0x10 

       0x00 ;; (item-hack) no-op
       0xE0 ;; load A into [FF00]
       0x00

       0xF0 ;; load 0xFF00 into A to get
       0x00 ;; button presses
       
       0xE6
       0x0F ;; select bottom four bits of A
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB8 ;; see if input is different (CP A B)

       0x00 ;; (item-hack) (INC SP)
       0x28 ;; repeat above steps if input is not different
       ;; (jump relative backwards if B != A)
       0xED ;; (literal -19) (item-hack) -19 == egg bomb (TM37)

       0x47 ;; load A into B
       
       0x0D ;; dec C
       0x37 ;; (item-hack) set-carry flag
       ;; branch based on C:
       0x20 ;; JR NZ
       23 ;; skip "input second nybble" and "jump to target" below
       
       ;; input second nybble

       0x0C ;; inc C
       0x0C ;; inc C

       0x00 ;; (item-hack) no-op
       0xE6 ;; select bottom bits
       0x0F
       0x37 ;; (item-hack) set-carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB2 ;; (OR A D) -> A

       0x22 ;; (do (A -> (HL)) (INC HL))

       0x1D ;; (DEC E)

       0x00 ;; (item-hack) 
       0x20 ;; jump back to input section if not done
       0xDA ;; literal -36 == TM 18 (counter)
       0x01 ;; (item-hack) set BC to literal (no-op)

       ;; jump to target
       0x00  ;; (item-hack) these two bytes can be anything.
       0x01 

       0x00   ;; (item-hack) no-op
       0xBF   ;; (CP A A) ensures Z
       
       0xCA   ;; (item-hack) jump if Z
       target-low
       target-high
       0x01   ;; (item-hack) will never be reached.
       
       ;; input first nybble
       0x00
       0xCB
       0x37  ;; swap nybbles on A

       0x57  ;; A -> D

       0x37  ;; (item-hack) set carry flag no-op
       0x18  ;; relative jump backwards
       0xCD  ;; literal -51 == TM05; go back to input section
       0x01  ;; (item-hack) will never reach this instruction

       ]
      (repeat 8 [0xFF 0x01])

      [;; jump to actual program
       0x00
       0x37  ;; (item-hack) set carry flag no-op

       0x2E  ;; 0x3A -> L
       0x3A


       0x00  ;; (item-hack) no-op
       0x26  ;; 0xD5 -> L
       0xD5  
       0x01  ;; (item-hack) set-carry BC

       0x00  ;; (item-hack) these can be anything
       0x01  

       0x00
       0xE9 ;; jump to (HL)
       ]])))

(defn view-desired-item-layout []
  (clojure.pprint/pprint
   (raw-inventory->inventory (pc-item-writer-program))))

(defn pc-item-writer-state []
    (-> (read-state "bootstrap-init")
      (set-memory pc-item-list-start 50)
      (set-memory-range
       map-function-address-start
       [0x8B 0xD5])
      (set-memory-range
       (inc pc-item-list-start)
       (pc-item-writer-program))))

(defn test-pc-item-writer []
  (let [orig (read-state "pc-item-writer")]
    (-> orig
    (print-listing 0xD162 (+ 0xD162 20))
    (run-moves (reduce concat
                (repeat 10 [[:a :b :start :select] []])))
    ((fn [_] (println "===========") _))
    (print-listing 0xD162 (+ 0xD162 20)))))

(defn close-all-menus [[moves state :as script]]
  (loop [s script]
    (let [depth (current-depth (second (do-nothing 50 s)))]
      (println "depth" depth)
      (if (= depth 1)
        s
        (recur (close-menu s))))))

(defn-memo name-rival
  ([] (name-rival (to-rival-name)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (first-difference [] [:r] DE)
          (play-moves
           [[]
            [] [] [:r] [] [:d] [:a]           ;; L
            [:r] [] [:r] [] [:r] [] [:r] [] 
            [:r] [] [:d] [] [:d] [:a]         ;; [PK]
            [:d] [] [:r] [:a]
            ]))))

(defn-memo to-room-pc
  ([] (to-room-pc (name-rival)))
  ([script]
     (->> script
          finish-title
          (walk [← ← ↑ ← ↑ ↑ ↑]))))

(defn-memo bootstrap-corrupt-save
  ([] (bootstrap-corrupt-save (to-room-pc)))
  ([script]
   (->> script
        (do-save-corruption 2)
        (corrupt-item-list 0)
        close-all-menus)))

(defn-memo begin-initial-deposits
  ([] (begin-initial-deposits
       (bootstrap-corrupt-save)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (scroll-text)
          (set-cursor 1)
          select-menu-entry)))

(defn deposit-held-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-item state n))]
    (println "total-quantity" total-quantity)
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         ((fn [[moves state :as script]]
            (if (not= (item-quantity-selected state) 1)
              (repeat-until-different [] item-quantity-selected script)
              script)))
         (set-quantity total-quantity quantity)
         (delayed-difference [] [:a] 100 #(search-string % "stored"))
         (scroll-text))))

(defn widthdraw-pc-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-pc-item state n))]
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         ((fn [[moves state :as script]]
            (if (not= (item-quantity-selected state) 1)
              (repeat-until-different [] item-quantity-selected
                                      script)
              script)))
         (set-quantity total-quantity quantity)
         (delayed-difference [] [:a] 100 #(search-string % "Withdrew"))
         (scroll-text))))

(defn toss-held-item [n quantity [moves state :as script]]
   (let [total-quantity (second (nth-item state n))]
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         (set-cursor-relative 1)
         (select-menu-entry -1)
         ((fn [[moves state :as script]]
            (if (not= (item-quantity-selected state) 1)
              (repeat-until-different [] item-quantity-selected
                                      script)
              script)))
         (set-quantity total-quantity quantity)
         (play-moves [[:a]])
         (scroll-text)
         (delayed-difference [] [:a] 100 #(search-string % "Threw"))
         (scroll-text)
         )))


(def desired-zero-quantities
  (map second (filter (comp (partial = 0) first)
                      (partition 2 (pc-item-writer-program)))))
  
(defn-memo initial-deposits
  ([] (initial-deposits (begin-initial-deposits)))
  ([script]
     (->> script
          (deposit-held-item 0 0xFE)
          (deposit-held-item 1 0xFE)
          ((fn [script]
             (reduce 
              (fn [script item] (deposit-held-item item 0xFF script))
              script
              (range 2 (+ 12 2)))))
          close-all-menus)))


(defn-memo prepare-celadon-warp
  ([] (prepare-celadon-warp (initial-deposits)))
  ([script]
     (->> script
          (activate-start-menu)
          (set-cursor-relative 1)
          (select-menu-entry)
          (toss-held-item 35 0xFA)
          (close-all-menus))))


;;0 -- 256
;;1 -- 254
;;2 -- 254
;;3 -- 255

(defn-memo restore-items
  ([] (restore-items (prepare-celadon-warp)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (scroll-text)
          (select-menu-entry)
          (widthdraw-pc-item 0 1)
          (widthdraw-pc-item 3 99)
          (widthdraw-pc-item 3 99)
          (widthdraw-pc-item 3 57)
          (widthdraw-pc-item 2 99)
          (widthdraw-pc-item 2 99)
          (widthdraw-pc-item 2 56)
          (widthdraw-pc-item 1 99)
          (widthdraw-pc-item 1 99)
          (widthdraw-pc-item 1 56)
          (widthdraw-pc-item 0 99)
          (widthdraw-pc-item 0 99)
          (widthdraw-pc-item 0 57)
          (close-all-menus))))

(defn-memo to-celadon
  ([] (to-celadon (restore-items)))
  ([script]
     (->> script
          (walk [→ → → → → → → ↑
                 ↓ ↓ ↓ ↓ ↓ ← ← ← ←
                 ↓ ↓]))))





          
        

